@@ -1,17 +1,32 @@
 /* Copyright (c) 2007 Timothy Wall, All Rights Reserved
  *
- * This library is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Lesser General Public
- * License as published by the Free Software Foundation; either
- * version 2.1 of the License, or (at your option) any later version.
+ * The contents of this file is dual-licensed under 2
+ * alternative Open Source/Free licenses: LGPL 2.1 or later and
+ * Apache License 2.0. (starting with JNA version 4.0.0).
  *
- * This library is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * Lesser General Public License for more details.
+ * You can freely decide which license you want to apply to
+ * the project.
+ *
+ * You may obtain a copy of the LGPL License at:
+ *
+ * http://www.gnu.org/licenses/licenses.html
+ *
+ * A copy is also included in the downloadable source code package
+ * containing JNA, in file "LGPL2.1".
+ *
+ * You may obtain a copy of the Apache License at:
+ *
+ * http://www.apache.org/licenses/
+ *
+ * A copy is also included in the downloadable source code package
+ * containing JNA, in file "AL2.0".
  */
 package com.sun.jna.platform.win32;
 
+import static com.sun.jna.platform.win32.WinBase.WAIT_OBJECT_0;
+import static com.sun.jna.platform.win32.WinNT.MEM_COMMIT;
+import static com.sun.jna.platform.win32.WinNT.MEM_RESERVE;
+import static com.sun.jna.platform.win32.WinNT.PAGE_EXECUTE_READWRITE;
 import static com.sun.jna.platform.win32.WinioctlUtil.FSCTL_GET_COMPRESSION;
 import static com.sun.jna.platform.win32.WinioctlUtil.FSCTL_GET_REPARSE_POINT;
 import static com.sun.jna.platform.win32.WinioctlUtil.FSCTL_SET_COMPRESSION;
@@ -36,11 +51,13 @@ import java.util.Collection;
 import java.util.Date;
 import java.util.List;
 import java.util.TimeZone;
+import java.util.concurrent.CountDownLatch;
+import java.util.concurrent.TimeUnit;
+
+import org.junit.Test;
 
-import com.sun.jna.Function;
 import com.sun.jna.Memory;
 import com.sun.jna.Native;
-import com.sun.jna.NativeLibrary;
 import com.sun.jna.NativeMappedConverter;
 import com.sun.jna.Platform;
 import com.sun.jna.Pointer;
@@ -59,6 +76,7 @@ import com.sun.jna.platform.win32.WinBase.SYSTEMTIME;
 import com.sun.jna.platform.win32.WinBase.SYSTEM_INFO;
 import com.sun.jna.platform.win32.WinBase.WIN32_FIND_DATA;
 import com.sun.jna.platform.win32.WinDef.DWORD;
+import com.sun.jna.platform.win32.WinDef.DWORDByReference;
 import com.sun.jna.platform.win32.WinDef.HMODULE;
 import com.sun.jna.platform.win32.WinDef.HWND;
 import com.sun.jna.platform.win32.WinDef.USHORT;
@@ -94,10 +112,12 @@ public class Kernel32Test extends TestCase {
     public void testNoDuplicateMethodsNames() {
         Collection<String> dupSet = AbstractWin32TestSupport.detectDuplicateMethods(Kernel32.class);
         if (dupSet.size() > 0) {
-            for (String name : new String[] {
-                    // has 2 overloads by design since the API accepts both OSVERSIONINFO and OSVERSIONINFOEX
-                    "GetVersionEx"
-                }) {
+            for (String name : new String[]{
+                // has 2 overloads by design since the API accepts both OSVERSIONINFO and OSVERSIONINFOEX
+                "GetVersionEx",
+                // one version is kind-of broken and retained for compatiblity (deprecated)
+                "CreateRemoteThread"
+            }) {
                 dupSet.remove(name);
             }
         }
@@ -274,6 +294,25 @@ public class Kernel32Test extends TestCase {
         }
     }
 
+    public void testOpenEvent() {
+        HANDLE handle = null, handle2 = null;
+
+        try {
+            handle = Kernel32.INSTANCE.CreateEvent(null, false, false, "jna-kernel32test");
+            assertNotNull("Failed to create event: " + Kernel32.INSTANCE.GetLastError(), handle);
+
+            handle2 = Kernel32.INSTANCE.OpenEvent(WinNT.EVENT_MODIFY_STATE, false, "jna-kernel32test");
+            assertNotNull("Failed to open event: " + Kernel32.INSTANCE.GetLastError(), handle2);
+
+            Kernel32.INSTANCE.SetEvent(handle2);
+
+            assertEquals(WinBase.WAIT_OBJECT_0, Kernel32.INSTANCE.WaitForSingleObject(handle, 1000));
+        } finally {
+            Kernel32Util.closeHandle(handle);
+            Kernel32Util.closeHandle(handle2);
+        }
+    }
+
     public void testResetEvent() {
         HANDLE handle = Kernel32.INSTANCE.CreateEvent(null, true, false, null);
         assertNotNull("Failed to create event: " + Kernel32.INSTANCE.GetLastError(), handle);
@@ -334,16 +373,14 @@ public class Kernel32Test extends TestCase {
     public void testGetCurrentThread() {
         HANDLE h = Kernel32.INSTANCE.GetCurrentThread();
         assertNotNull("No current thread handle", h);
-        assertFalse("Null current thread handle", h.equals(0));
-        // Calling the CloseHandle function with this handle has no effect
-        Kernel32Util.closeHandle(h);
+        assertNotNull("Null current thread handle", h.getPointer());
     }
 
     public void testOpenThread() {
         HANDLE h = Kernel32.INSTANCE.OpenThread(WinNT.THREAD_ALL_ACCESS, false,
                 Kernel32.INSTANCE.GetCurrentThreadId());
         assertNotNull(h);
-        assertFalse(h.equals(0));
+        assertNotNull(h.getPointer());
         Kernel32Util.closeHandle(h);
     }
 
@@ -354,9 +391,7 @@ public class Kernel32Test extends TestCase {
     public void testGetCurrentProcess() {
         HANDLE h = Kernel32.INSTANCE.GetCurrentProcess();
         assertNotNull("No current process handle", h);
-        assertFalse("Null current process handle", h.equals(0));
-        // Calling the CloseHandle function with a pseudo handle has no effect
-        Kernel32Util.closeHandle(h);
+        assertNotNull("Null current process handle", h.getPointer());
     }
 
     public void testOpenProcess() {
@@ -383,6 +418,51 @@ public class Kernel32Test extends TestCase {
         }
     }
 
+    public void testGetProcessTimes() {
+        int pid = Kernel32.INSTANCE.GetCurrentProcessId();
+        HANDLE h = Kernel32.INSTANCE.OpenProcess(WinNT.PROCESS_QUERY_INFORMATION, false, pid);
+        assertNotNull("Failed (" + Kernel32.INSTANCE.GetLastError() + ") to get process ID=" + pid + " handle", h);
+
+        try {
+            FILETIME lpCreationTime = new FILETIME();
+            FILETIME lpExitTime = new FILETIME();
+            FILETIME lpKernelTime = new FILETIME();
+            FILETIME lpUserTime = new FILETIME();
+            boolean b = Kernel32.INSTANCE.GetProcessTimes(h, lpCreationTime, lpExitTime, lpKernelTime, lpUserTime);
+            assertTrue("Failed (" + Kernel32.INSTANCE.GetLastError() + ") to get process times", b);
+            // Process must have started before now.
+            long upTimeMillis = System.currentTimeMillis() - lpCreationTime.toTime();
+            assertTrue(upTimeMillis >= 0);
+            // lpExitTime is undefined for a running process, do not test
+            // Kernel and User time must be < up time (in 100ns ticks)
+            assertTrue(lpKernelTime.toDWordLong().longValue()
+                    + lpUserTime.toDWordLong().longValue() < (upTimeMillis + 1L) * 10000L);
+        } finally {
+            Kernel32Util.closeHandle(h);
+        }
+    }
+
+    public void testGetProcessIoCounters() {
+        int pid = Kernel32.INSTANCE.GetCurrentProcessId();
+        HANDLE h = Kernel32.INSTANCE.OpenProcess(WinNT.PROCESS_QUERY_INFORMATION, false, pid);
+        assertNotNull("Failed (" + Kernel32.INSTANCE.GetLastError() + ") to get process ID=" + pid + " handle", h);
+
+        try {
+            WinNT.IO_COUNTERS lpIoCounters = new WinNT.IO_COUNTERS();
+            boolean b = Kernel32.INSTANCE.GetProcessIoCounters(h, lpIoCounters);
+            assertTrue("Failed (" + Kernel32.INSTANCE.GetLastError() + ") to get process IO counters", b);
+            // IO must be nonzero
+            assertTrue(lpIoCounters.ReadOperationCount >= 0);
+            assertTrue(lpIoCounters.WriteOperationCount >= 0);
+            assertTrue(lpIoCounters.OtherOperationCount >= 0);
+            assertTrue(lpIoCounters.ReadTransferCount >= 0);
+            assertTrue(lpIoCounters.WriteTransferCount >= 0);
+            assertTrue(lpIoCounters.OtherTransferCount >= 0);
+        } finally {
+            Kernel32Util.closeHandle(h);
+        }
+    }
+
     public void testGetTempPath() {
         char[] buffer = new char[WinDef.MAX_PATH];
         assertTrue(Kernel32.INSTANCE.GetTempPath(new DWORD(WinDef.MAX_PATH), buffer).intValue() > 0);
@@ -444,22 +524,26 @@ public class Kernel32Test extends TestCase {
         SYSTEM_INFO lpSystemInfo = new SYSTEM_INFO();
         Kernel32.INSTANCE.GetSystemInfo(lpSystemInfo);
         assertTrue(lpSystemInfo.dwNumberOfProcessors.intValue() > 0);
+        // the dwOemID member is obsolete, but gets a value.
+        // the pi member is a structure and isn't read by default
+        assertEquals(lpSystemInfo.processorArchitecture.dwOemID.getLow(),
+                lpSystemInfo.processorArchitecture.pi.wProcessorArchitecture);
     }
 
     public void testGetSystemTimes() {
-      Kernel32 kernel = Kernel32.INSTANCE;
-      FILETIME lpIdleTime = new FILETIME();
-      FILETIME lpKernelTime = new FILETIME();
-      FILETIME lpUserTime = new FILETIME();
-      boolean succ = kernel.GetSystemTimes(lpIdleTime, lpKernelTime, lpUserTime);
-      assertTrue(succ);
-      long idleTime = lpIdleTime.toDWordLong().longValue();
-      long kernelTime = lpKernelTime.toDWordLong().longValue();
-      long userTime = lpUserTime.toDWordLong().longValue();
-      // All should be >= 0.  kernel includes idle.
-      assertTrue(idleTime >= 0);
-      assertTrue(kernelTime >= idleTime);
-      assertTrue(userTime >= 0);
+        Kernel32 kernel = Kernel32.INSTANCE;
+        FILETIME lpIdleTime = new FILETIME();
+        FILETIME lpKernelTime = new FILETIME();
+        FILETIME lpUserTime = new FILETIME();
+        boolean succ = kernel.GetSystemTimes(lpIdleTime, lpKernelTime, lpUserTime);
+        assertTrue(succ);
+        long idleTime = lpIdleTime.toDWordLong().longValue();
+        long kernelTime = lpKernelTime.toDWordLong().longValue();
+        long userTime = lpUserTime.toDWordLong().longValue();
+        // All should be >= 0.  kernel includes idle.
+        assertTrue(idleTime >= 0);
+        assertTrue(kernelTime >= idleTime);
+        assertTrue(userTime >= 0);
     }
 
     public void testIsWow64Process() {
@@ -1219,12 +1303,45 @@ public class Kernel32Test extends TestCase {
         }
     }
 
-    public final void testCreateRemoteThread() throws IOException {
-        HANDLE hThrd = Kernel32.INSTANCE.CreateRemoteThread(null, null, 0, null, null, null, null);
+    public final void testCreateRemoteThreadInvalid() throws IOException {
+        HANDLE hThrd = Kernel32.INSTANCE.CreateRemoteThread(null, null, 0, (Pointer) null, null, 0, null);
         assertNull(hThrd);
         assertEquals(Kernel32.INSTANCE.GetLastError(), WinError.ERROR_INVALID_HANDLE);
     }
 
+    @Test
+    public void testCreateRemoteThread() {
+        Pointer addr = Kernel32.INSTANCE.VirtualAllocEx(
+            Kernel32.INSTANCE.GetCurrentProcess(), null, new SIZE_T(4096),
+            MEM_COMMIT | MEM_RESERVE, PAGE_EXECUTE_READWRITE);
+
+        // mov eax, ecx; ret; int3
+        Memory localBuffer = new Memory(4096);
+        localBuffer.setInt(0, 0xccc3c18b);
+        IntByReference bytesWritten = new IntByReference();
+        Kernel32.INSTANCE.WriteProcessMemory(Kernel32.INSTANCE.GetCurrentProcess(),
+            addr, localBuffer, 4096, bytesWritten);
+        assertEquals(4096, bytesWritten.getValue());
+
+        DWORDByReference threadId = new DWORDByReference();
+        HANDLE hThread = Kernel32.INSTANCE.CreateRemoteThread(
+            Kernel32.INSTANCE.GetCurrentProcess(),
+            null, 0, addr, new Pointer(12345), 0, threadId);
+        assertNotNull(hThread);
+        assertTrue(threadId.getValue().longValue() > 0);
+
+        int waitResult = Kernel32.INSTANCE.WaitForSingleObject(hThread, 10000);
+        assertEquals(WAIT_OBJECT_0, waitResult);
+
+        IntByReference exitCode = new IntByReference();
+        boolean exitResult = Kernel32.INSTANCE.GetExitCodeThread(hThread, exitCode);
+        assertTrue(exitResult);
+        assertEquals(12345, exitCode.getValue());
+
+        assertTrue(Kernel32.INSTANCE.VirtualFreeEx(Kernel32.INSTANCE.GetCurrentProcess(),
+            addr, new SIZE_T(0), WinNT.MEM_RELEASE));
+    }
+
     public void testWriteProcessMemory() {
         Kernel32 kernel = Kernel32.INSTANCE;
 
@@ -1314,7 +1431,7 @@ public class Kernel32Test extends TestCase {
                     case WinBase.CBR_56000:
                     case WinBase.CBR_600:
                     case WinBase.CBR_9600:
-                    break;
+                        break;
                     default:
                         fail("Received value of WinBase.DCB.BaudRate is not valid");
                 }
@@ -1598,7 +1715,7 @@ public class Kernel32Test extends TestCase {
             }
         }
     }
-    
+
     public void testSetErrorMode() {
         // Set bit flags to 0x0001
         int previousMode = Kernel32.INSTANCE.SetErrorMode(0x0001);
@@ -1612,45 +1729,96 @@ public class Kernel32Test extends TestCase {
 //    /**
 //     * Test that a named function on win32 can be equally resolved by its ordinal
 //     * value.
-//     * 
+//     *
 //     * From link.exe /dump /exports c:\\Windows\\System32\\kernel32.dll
-//     * 
+//     *
 //     *  746  2E9 0004FA20 GetTapeStatus
 //     *  747  2EA 0002DB20 GetTempFileNameA
 //     *  748  2EB 0002DB30 GetTempFileNameW
 //     *  749  2EC 0002DB40 GetTempPathA
 //     *  750  2ED 0002DB50 GetTempPathW
 //     *  751  2EE 00026780 GetThreadContext
-//     * 
+//     *
 //     * The tested function is GetTempPathW which is mapped to the ordinal 750.
 //     */
 //    public void testGetProcAddress() {
 //        NativeLibrary kernel32Library = NativeLibrary.getInstance("kernel32");
 //        // get module handle needed to resolve function pointer via GetProcAddress
 //        HMODULE kernel32Module = Kernel32.INSTANCE.GetModuleHandle("kernel32");
-//        
+//
 //        Function namedFunction = kernel32Library.getFunction("GetTempPathW");
 //        long namedFunctionPointerValue = Pointer.nativeValue(namedFunction);
-//        
+//
 //        Pointer ordinalFunction = Kernel32.INSTANCE.GetProcAddress(kernel32Module, 750);
 //        long ordinalFunctionPointerValue = Pointer.nativeValue(ordinalFunction);
-//        
+//
 //        assertEquals(namedFunctionPointerValue, ordinalFunctionPointerValue);
 //    }
-    
+
     public void testSetThreadExecutionState() {
         int originalExecutionState = Kernel32.INSTANCE.SetThreadExecutionState(
                 WinBase.ES_CONTINUOUS | WinBase.ES_SYSTEM_REQUIRED | WinBase.ES_AWAYMODE_REQUIRED
         );
-        
+
         assert originalExecutionState > 0;
-        
+
         int intermediateExecutionState = Kernel32.INSTANCE.SetThreadExecutionState(
                 WinBase.ES_CONTINUOUS
         );
-        
+
         assertEquals(WinBase.ES_CONTINUOUS | WinBase.ES_SYSTEM_REQUIRED | WinBase.ES_AWAYMODE_REQUIRED, intermediateExecutionState);
-        
+
         Kernel32.INSTANCE.SetThreadExecutionState(originalExecutionState);
     }
+
+    public void testMutex() throws InterruptedException {
+        HANDLE mutexHandle = Kernel32.INSTANCE.CreateMutex(null, true, "JNA-Test-Mutex");
+
+        assertNotNull(mutexHandle);
+
+        final CountDownLatch preWait = new CountDownLatch(1);
+        final CountDownLatch postWait = new CountDownLatch(1);
+        final CountDownLatch postRelease = new CountDownLatch(1);
+
+        final Exception[] exceptions = new Exception[1];
+        Thread t = new Thread() {
+            @Override
+            public void run() {
+                try {
+                    HANDLE mutexHandle2 = Kernel32.INSTANCE.OpenMutex(WinNT.SYNCHRONIZE, false, "JNA-Test-Mutex");
+                    try {
+                        assertNotNull(mutexHandle2);
+                        preWait.countDown();
+                        int result = Kernel32.INSTANCE.WaitForSingleObject(mutexHandle2, WinBase.INFINITE);
+                        assertEquals(result, WinBase.WAIT_OBJECT_0);
+                        postWait.countDown();
+                    } finally {
+                        Kernel32.INSTANCE.ReleaseMutex(mutexHandle2);
+                        Kernel32.INSTANCE.CloseHandle(mutexHandle2);
+                        postRelease.countDown();
+                    }
+                } catch (Exception ex) {
+                    exceptions[0] = ex;
+                }
+            }
+        };
+
+        t.start();
+
+        assertTrue(preWait.await(2, TimeUnit.SECONDS));
+
+        Kernel32.INSTANCE.ReleaseMutex(mutexHandle);
+
+        assertTrue(postWait.await(2, TimeUnit.SECONDS));
+
+        Kernel32.INSTANCE.CloseHandle(mutexHandle);
+
+        assertTrue(postRelease.await(2, TimeUnit.SECONDS));
+
+        assertNull(exceptions[0]);
+
+        mutexHandle = Kernel32.INSTANCE.OpenMutex(WinNT.SYNCHRONIZE, false, "JNA-Test-Mutex");
+
+        assertNull(mutexHandle);
+    }
 }
