@@ -1,14 +1,25 @@
 /* Copyright (c) 2010, 2013 Daniel Doubrovkine, Markus Karg, All Rights Reserved
  *
- * This library is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Lesser General Public
- * License as published by the Free Software Foundation; either
- * version 2.1 of the License, or (at your option) any later version.
+ * The contents of this file is dual-licensed under 2
+ * alternative Open Source/Free licenses: LGPL 2.1 or later and
+ * Apache License 2.0. (starting with JNA version 4.0.0).
  *
- * This library is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * Lesser General Public License for more details.
+ * You can freely decide which license you want to apply to
+ * the project.
+ *
+ * You may obtain a copy of the LGPL License at:
+ *
+ * http://www.gnu.org/licenses/licenses.html
+ *
+ * A copy is also included in the downloadable source code package
+ * containing JNA, in file "LGPL2.1".
+ *
+ * You may obtain a copy of the Apache License at:
+ *
+ * http://www.apache.org/licenses/
+ *
+ * A copy is also included in the downloadable source code package
+ * containing JNA, in file "AL2.0".
  */
 package com.sun.jna.platform.win32;
 
@@ -19,15 +30,24 @@ import java.io.FileReader;
 import java.io.FileWriter;
 import java.io.IOException;
 import java.io.PrintWriter;
+import java.util.ArrayList;
 import java.util.Collection;
 import java.util.List;
 import java.util.Map;
 
 import com.sun.jna.Pointer;
 import com.sun.jna.platform.win32.Tlhelp32.MODULEENTRY32W;
+import com.sun.jna.platform.win32.WinNT.CACHE_RELATIONSHIP;
+import com.sun.jna.platform.win32.WinNT.GROUP_RELATIONSHIP;
 import com.sun.jna.platform.win32.WinNT.HANDLE;
 import com.sun.jna.platform.win32.WinNT.HRESULT;
 import com.sun.jna.platform.win32.WinNT.LARGE_INTEGER;
+import com.sun.jna.platform.win32.WinNT.LOGICAL_PROCESSOR_RELATIONSHIP;
+import com.sun.jna.platform.win32.WinNT.NUMA_NODE_RELATIONSHIP;
+import com.sun.jna.platform.win32.WinNT.PROCESSOR_CACHE_TYPE;
+import com.sun.jna.platform.win32.WinNT.PROCESSOR_RELATIONSHIP;
+import com.sun.jna.platform.win32.WinNT.SYSTEM_LOGICAL_PROCESSOR_INFORMATION;
+import com.sun.jna.platform.win32.WinNT.SYSTEM_LOGICAL_PROCESSOR_INFORMATION_EX;
 
 import junit.framework.TestCase;
 
@@ -342,4 +362,120 @@ public class Kernel32UtilTest extends TestCase {
         // so assert total count is at least two
         assertTrue("This is supposed to return all the modules in a process, so there should be an EXE and at least 1 Windows API DLL.", results.size() > 2);
     }
+
+    public void testExpandEnvironmentStrings() {
+        Kernel32.INSTANCE.SetEnvironmentVariable("DemoVariable", "DemoValue");
+        assertEquals("DemoValue", Kernel32Util.expandEnvironmentStrings("%DemoVariable%"));
+    }
+
+    public void testGetLogicalProcessorInformation() {
+        SYSTEM_LOGICAL_PROCESSOR_INFORMATION[] procInfo = Kernel32Util
+                .getLogicalProcessorInformation();
+        assertTrue(procInfo.length > 0);
+    }
+
+    public void testGetLogicalProcessorInformationEx() {
+        SYSTEM_LOGICAL_PROCESSOR_INFORMATION_EX[] procInfo = Kernel32Util
+                .getLogicalProcessorInformationEx(WinNT.LOGICAL_PROCESSOR_RELATIONSHIP.RelationAll);
+        List<GROUP_RELATIONSHIP> groups = new ArrayList<GROUP_RELATIONSHIP>();
+        List<PROCESSOR_RELATIONSHIP> packages = new ArrayList<PROCESSOR_RELATIONSHIP>();
+        List<NUMA_NODE_RELATIONSHIP> numaNodes = new ArrayList<NUMA_NODE_RELATIONSHIP>();
+        List<CACHE_RELATIONSHIP> caches = new ArrayList<CACHE_RELATIONSHIP>();
+        List<PROCESSOR_RELATIONSHIP> cores = new ArrayList<PROCESSOR_RELATIONSHIP>();
+
+        for (int i = 0; i < procInfo.length; i++) {
+            // Build list from relationship
+            switch (procInfo[i].relationship) {
+                case LOGICAL_PROCESSOR_RELATIONSHIP.RelationGroup:
+                    groups.add((GROUP_RELATIONSHIP) procInfo[i]);
+                    break;
+                case LOGICAL_PROCESSOR_RELATIONSHIP.RelationProcessorPackage:
+                    packages.add((PROCESSOR_RELATIONSHIP) procInfo[i]);
+                    break;
+                case LOGICAL_PROCESSOR_RELATIONSHIP.RelationNumaNode:
+                    numaNodes.add((NUMA_NODE_RELATIONSHIP) procInfo[i]);
+                    break;
+                case LOGICAL_PROCESSOR_RELATIONSHIP.RelationCache:
+                    caches.add((CACHE_RELATIONSHIP) procInfo[i]);
+                    break;
+                case LOGICAL_PROCESSOR_RELATIONSHIP.RelationProcessorCore:
+                    cores.add((PROCESSOR_RELATIONSHIP) procInfo[i]);
+                    break;
+                default:
+                    throw new IllegalStateException("Unmapped relationship.");
+            }
+            // Test that native provided size matches JNA structure size
+            assertEquals(procInfo[i].size, procInfo[i].size());
+        }
+
+        // Test that getting all relations matches the same totals as
+        // individuals.
+        assertEquals(groups.size(), Kernel32Util
+                .getLogicalProcessorInformationEx(WinNT.LOGICAL_PROCESSOR_RELATIONSHIP.RelationGroup).length);
+        assertEquals(packages.size(), Kernel32Util.getLogicalProcessorInformationEx(
+                WinNT.LOGICAL_PROCESSOR_RELATIONSHIP.RelationProcessorPackage).length);
+        assertEquals(numaNodes.size(), Kernel32Util
+                .getLogicalProcessorInformationEx(WinNT.LOGICAL_PROCESSOR_RELATIONSHIP.RelationNumaNode).length);
+        assertEquals(caches.size(), Kernel32Util
+                .getLogicalProcessorInformationEx(WinNT.LOGICAL_PROCESSOR_RELATIONSHIP.RelationCache).length);
+        assertEquals(cores.size(), Kernel32Util
+                .getLogicalProcessorInformationEx(WinNT.LOGICAL_PROCESSOR_RELATIONSHIP.RelationProcessorCore).length);
+
+        // Test GROUP_RELATIONSHIP
+        assertEquals(1, groups.size()); // Should only be one group structure
+        for (GROUP_RELATIONSHIP group : groups) {
+            assertEquals(LOGICAL_PROCESSOR_RELATIONSHIP.RelationGroup, group.relationship);
+            assertTrue(group.activeGroupCount <= group.maximumGroupCount);
+            assertEquals(group.activeGroupCount, group.groupInfo.length);
+            for (int j = 0; j < group.activeGroupCount; j++) {
+                assertTrue(group.groupInfo[j].activeProcessorCount <= group.groupInfo[j].maximumProcessorCount);
+                assertEquals(group.groupInfo[j].activeProcessorCount,
+                        Long.bitCount(group.groupInfo[j].activeProcessorMask.longValue()));
+                assertTrue(group.groupInfo[j].maximumProcessorCount <= 64);
+            }
+        }
+
+        // Test PROCESSOR_RELATIONSHIP packages
+        assertTrue(cores.size() >= packages.size());
+        for (PROCESSOR_RELATIONSHIP pkg : packages) {
+            assertEquals(LOGICAL_PROCESSOR_RELATIONSHIP.RelationProcessorPackage, pkg.relationship);
+            assertEquals(0, pkg.flags); // packages have 0 flags
+            assertEquals(0, pkg.efficiencyClass); // packages have 0 efficiency
+            assertEquals(pkg.groupCount, pkg.groupMask.length);
+        }
+
+        // Test PROCESSOR_RELATIONSHIP cores
+        for (PROCESSOR_RELATIONSHIP core : cores) {
+            assertEquals(LOGICAL_PROCESSOR_RELATIONSHIP.RelationProcessorCore, core.relationship);
+            // Hyperthreading flag set if at least 2 logical processors
+            assertTrue(Long.bitCount(core.groupMask[0].mask.longValue()) > 0);
+            if (Long.bitCount(core.groupMask[0].mask.longValue()) > 1) {
+                assertEquals(WinNT.LTP_PC_SMT, core.flags);
+            } else {
+                assertEquals(0, core.flags);
+            }
+            // Cores are always in one group
+            assertEquals(1, core.groupCount);
+            assertEquals(1, core.groupMask.length);
+        }
+
+        // Test NUMA_NODE_RELATIONSHIP
+        for (NUMA_NODE_RELATIONSHIP numaNode : numaNodes) {
+            assertEquals(LOGICAL_PROCESSOR_RELATIONSHIP.RelationNumaNode, numaNode.relationship);
+            assertTrue(numaNode.nodeNumber >= 0);
+        }
+
+        // Test CACHE_RELATIONSHIP
+        for (CACHE_RELATIONSHIP cache : caches) {
+            assertEquals(LOGICAL_PROCESSOR_RELATIONSHIP.RelationCache, cache.relationship);
+            assertTrue(cache.level >= 1);
+            assertTrue(cache.level <= 4);
+            assertTrue(cache.cacheSize > 0);
+            assertTrue(cache.lineSize > 0);
+            assertTrue(cache.type == PROCESSOR_CACHE_TYPE.CacheUnified
+                    || cache.type == PROCESSOR_CACHE_TYPE.CacheInstruction
+                    || cache.type == PROCESSOR_CACHE_TYPE.CacheData || cache.type == PROCESSOR_CACHE_TYPE.CacheTrace);
+            assertTrue(cache.associativity == WinNT.CACHE_FULLY_ASSOCIATIVE || cache.associativity > 0);
+        }
+    }
 }
